
% Test-case of dense water overflow.
% See Ilicak et al. 2012, Ocean Modelling, vol.45-46, p.37-58,
%     https://doi.org/10.1016/j.ocemod.2011.10.003.
% (see also Haidvogel & Beckmann 1999 who discuss a similar test-case.)

  clear;
  close all;
  more  off;

% Path to input/output files (must end with character '/'):
  outd = '/tmp/';
  nlay = 2;                    % Number  of LAYers.
  rhon = [1028., 1030.];       % Density of layers (kg/m3).
  dl   = 1.e3;                 % Mesh size (meters).
  lm   = round( 200.e3 / dl ); % Grid size in along -stream (x) direction.
  mm   = 1;                    % Grid size in across-stream (y) direction.

% IndeX to the center of the model domain, in the y direction:
  ix_0 = ceil(0.5 * (mm + 2));

% x,y positions corresponding to the center of each grid cell:
  [xx, yy] = meshgrid( (1 : lm + 2) - 1.5, ...
                       (1 : mm + 2) - 1.5 );
  xx   = xx' * dl; clear yy; % Units: meters.

% Define BOttom topography of modal domain (increasing downward).
%   Mimic the domain of Ilicak et al. 2012.
  h_bo = 0.5 * (1. + tanh( (xx - 40.e3) / 7.e3 ));
  h_bo = h_bo * 1500. + 500.; % Units: meters.

% Bottom SLOPe in the x direction:
  slop = zeros( lm + 2, 1 );
  slop( 2 : end - 1 ) = h_bo( 3 : end, ix_0 ) - h_bo( 1 : end - 2, ix_0 );
  slop = abs( slop ) / (2. * dl);

% Margins of bottom topography should always be set to zero in beom:
  h_bo( 1,  : ) = 0.;
  h_bo(end, : ) = 0.;
  h_bo( :,  1 ) = 0.;
  h_bo( :, end) = 0.;

% Compute the SALmon's thickness `hsal' (see the discussion about this test-case
%   in beom's documentation).
  hsal = 50. * max(slop(:)) * dl * (rhon(2) - rhon(1)) / rhon(2); % meters.

% MINimum layer thickness allowed in the calculation. beom always assumes a
%   factor of 10 between parameters hsal and hmin:
  hmin = hsal / 10.;

  ndeg = get_nbr_deg_freedom( h_bo ); % Number of DEGrees of freedom.
  grav = 9.8;                         % GRAVity (m/s2).
  cext = sqrt(grav * max(h_bo(:)));   % Celerity of surface gravity waves (m/s).
  clear grav;

% Vertical position for the TOP of the Layers in the state of rest and at the
%   deepest point of the model domain. The top of layer 1 coincides with the
%   geoid (0.) while layer 2 is positioned one Salmon's thickness above the
%   seabed. topl is positive downward and non-dimensionalized by the maximum
%   bottom depth of the model domain.
  topl = [0., 1. - hsal / max( h_bo(:) )]; % Non-dimensional.

% The initial condition has u=v=0 and a mass of dense water in the first 20km
%   of the domain (see Ilicak et al.)
  n    = zeros( lm + 2, mm + 2, nlay );
  u    = zeros( lm + 2, mm + 2, nlay );
  v    = zeros( lm + 2, mm + 2, nlay );

% Within the first 20km, raise the interface of layer 2 by 500m minus 5xhsal.
%   The 5xhsal ensures that layer 1 has a thickness sufficient that the
%   artificial term of Salmon 2002 isn't playing a role in the dynamics.
  i20k = 1 + round( 20.e3 / dl );
  n( 2 : i20k, ix_0, 2 ) = 500. - 5. * hsal; clear i20k;

% The cascading dense water eventually reaches the eastern boundary at great
%   speed. Allow the cascade to exit the domain gracefully by defining a
%   nudging zone there.
  east = zeros( lm + 2, mm + 2, 3 ); % 3 variables: eta, u, v.
  npts =  15;                        % Number of grid PoinTS in nudging zone.
  widt = npts * dl;                  % WIDTh of nudging zone (meters).
  dt   = 0.5  * dl / cext;           % Model timestep (seconds).

  for i = 1 : lm + 2 % Eastern boundary.
%  `xpos' is the position within the sponge zone (in units of grid points).
    xpos        = i - 1.5 + npts - lm;
    xpos        = max([ xpos;        0.  ]);
    xpos        = min([ xpos; npts - 0.5 ]);
%   See Lavelle and Thacker 2008, Ocean Modelling, vol.20 p.270-292,
%       and Eq.29 of Modave et al. 2010 Ocean Dynamics vol.60 p.65-79.
%   Only nudge eta (1) and u (2); not v (3).
    east( i, :, [1, 2] ) = dt * cext / widt * xpos / (npts - xpos);
  end; clear i xpos npts widt dt;

% Write input files over disk. Before doing so, make sure that the directory
%   `outd' is clean (i.e., delete all input files inside `outd').

  if ( exist([ outd 'init.bin' ]) ); delete([ outd 'init.bin' ]); end
  if ( exist([ outd 'h_bo.bin' ]) ); delete([ outd 'h_bo.bin' ]); end
  if ( exist([ outd 'fcor.bin' ]) ); delete([ outd 'fcor.bin' ]); end
  if ( exist([ outd 'taus.bin' ]) ); delete([ outd 'taus.bin' ]); end
  if ( exist([ outd 'bodf.bin' ]) ); delete([ outd 'bodf.bin' ]); end
  if ( exist([ outd 'nudg.bin' ]) ); delete([ outd 'nudg.bin' ]); end
  if ( exist([ outd 'hdot.bin' ]) ); delete([ outd 'hdot.bin' ]); end
  if ( exist([ outd 'tide.bin' ]) ); delete([ outd 'tide.bin' ]); end

  [f_id, mesg] = fopen([ outd 'h_bo.bin' ],      'w', 'ieee-le' );
  cont = fwrite( f_id, h_bo,            'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

  [f_id, mesg] = fopen([ outd 'nudg.bin' ],      'w', 'ieee-le' );
  cont = fwrite( f_id, east,            'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont east;

  [f_id, mesg] = fopen([ outd 'init.bin' ],      'w', 'ieee-le' );
  cont = fwrite( f_id, cat(4, n, u, v), 'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont n u v;

% print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, dt_s, ...
%               dt_o, dt_r, dt3d, bvis, dvis, bdrg, hmin, hsbl, hbbl, g_fb, ...
%               uadv, qdrg, ocrp, rsta, xper, yper, diag, tauw, idir, odir, ...
%               desc );
  print_params( lm,   mm,   nlay, ndeg, dl,   cext, 0.,   rhon, topl, 1.,   ...
                1./96,0.,   0.,   0.,   0.5,  0.,   hmin, 10.,  10.,  1.,   ...
                1.,   0.,   1.,   0,    0.,   0.,   0.,   [0,0],outd, outd, ...
                'Test-case: Dense water overflow' );

% See beom documentation on how to compile and execute the program.

  disp(['Now compile and execute the program in a different terminal. ' ...
        'Press a key when it is over.']);
  pause

  disp('Reading the output data, please wait...');

% Read the calculation's metadata:

  [taxi, h_0, f0, dl, rhon, desc] = get_metadata( strtrim( outd ) );
  nrec = length( taxi ); % Number of RECords in the output files.

% X-axis for the filled polygons representing the two layers. Ignore the first
%   and last points in the x dimension (margins of model domain).
  xpol = [          xx( 2 : end - 1, ix_0 );      ...
           flipdim( xx( 2 : end - 1, ix_0 ), 1 ); ...
                    xx( 2,           ix_0 ) ] * 1.e-3; % Units of km.
  figure;
  for irec = 1 : nrec

%   Depth of LAYers' interface, positive downward, relative to the geoid (z=0).
    dlay = nan( lm + 2, mm + 2, nlay );
    n    = get_field( 'eta_', irec, outd );

    for ilay = 1 : nlay
      dlay(:,:,ilay) = h_bo                                      ...
                     - squeeze( sum( h_0(:,:,ilay : nlay), 3 ) ) ...
                     - n(:,:,ilay);
    end; clear ilay n;

%   Display the two layers (and the topography) as filled polygons:
    hand = fill( xpol,                                                ...
                 [         dlay( 2 : end - 1, ix_0, 1 );              ...
                  flipdim( dlay( 2 : end - 1, ix_0, 2 ), 1 );         ...
                           dlay( 2,           ix_0, 1 ) ], [0.5 0 0], ...
                 xpol,                                                ...
                 [         dlay( 2 : end - 1, ix_0, 2 );              ...
                  flipdim( h_bo( 2 : end - 1, ix_0    ), 1 );         ...
                           dlay( 2,           ix_0, 2 ) ], [0 0 0.5], ...
                 xpol,                                                ...
                 [         h_bo( 2 : end - 1, ix_0    );              ...
                  (max( h_bo(:) ) + 1.) * ones( lm, 1 );              ...
                           h_bo( 2,           ix_0    ) ], 0.6 * [1 1 1] );
    set( hand, 'edgecolor', 'none' ); clear hand;
    axis ij;
    set( gca, 'ylim', [- 1., max( h_bo(:) )], 'xtick', (0. : 20. : 200.) );
    ylabel( 'Depth (m)'     );
    xlabel( 'Distance (km)' );
    title([ 'Dense water overflow experiment, t = ' ...
            num2str( round( taxi( irec ) * 24. ), '%02u' ) ' hours' ]);

    pause( 0.01 );
  end; clear irec;

