
% Test-case of double gyre wind stress in a flat two-layer ocean.
% See Salmon (2002), p.628-630, https://doi.org/10.1357/002224002762324194

  clear;
  close all;
  more  off;

% Path to input/output files (must end with character '/'):
  outd = '/tmp/';
  vide = 0;                        % If vide=1, create a video.
  ebud = 0;                        % If ebud=1, build an energy budget.
  nlay = 2;                        % Number  of LAYers.
  rhon = 1030. * [1. - 2.e-3, 1.]; % Density of layers (kg/m3).
  lx   = 4000.e3;                  % Domain length in zonal direction (m).
  ly   = 8000.e3;                  % Domain length in merid.direction (m).
  lm   = 100;                      % Grid size     in zonal direction.
  dl   = round( lx / lm );         % Mesh size (m).
  mm   = round( ly / dl );         % Grid size     in merid.direction.
  hfla = 4000.;                    % Depth of the flat-bottomed ocean (m).
  dt_o = 2.;                       % Time between Outputs (days).
% Distribute wind stress over a `surface boundary layer' of a given thickness:
  hsbl = 100.;                     % (m), see Salmon 2002.

% x,y positions corresponding to the center of each grid cell:
  [xx, yy] = meshgrid( (1 : lm + 2) - 1.5, ...
                       (1 : mm + 2) - 1.5 );
  xx   = xx' * dl; % Units: meters.
  yy   = yy' * dl; % Units: meters.

% Assume a beta-plane f = beta y, and compute the central latitude:
  lat0 = 0.5 * ly / (0.5 * pi * 6371.e3) * 90.; % Degrees North.
  Omeg = 2. * pi / (24. * 3600.);               % Earth's rotation (rad/s).
% Compute the beta parameter (df/dy) at the central latitude:
  beta = 2. * Omeg * cosd( lat0 ) / 6371.e3;    % (rad/(s m)).
  clear lat0 Omeg;

  fcor = yy * beta;

% Salmon (2002): 2 dynes cm^(-2) <=> 0.2 Pascals.
  tauw = zeros( lm + 2, mm + 2, 2 );
  tauw(:, :, 1) = 0.2 * cos( 2. * pi * yy / (ly / 0.75) + pi );

% No gradient in wind stress across the edge of the model domain:
  tauw( 1,   :,  :) = tauw( 2,      :,       :);
  tauw( :,   1,  :) = tauw( :,      2,       :);
  tauw(end,  :,  :) = tauw(end - 1, :,       :);
  tauw( :,  end, :) = tauw( :,      end - 1, :);

% Salmon (2002) uses a Rayleigh drag of the form du/dt=... - lambda u.
%   This drag is applied to *both* layers (bottom and surface). There is no
%   easy way to mimic this behavior in beom (without hardcoding the term).
% Therefore, we replace Salmon's Rayleigh drag by a dynamic viscosity:
  dvis = 0.1;

% Vertical position for the TOP of the Layers in the state of rest:
  topl = [ 0., 100. / hfla ];      % (Non-dimensionalized by max. depth).

% MINimum layer thickness allowed in the calculation. beom always assumes a
%   factor of 10 between parameters hsal and hmin:
  hmin = 20. / 10.;                % Salmon (2002) uses hsal=20m.

% Bottom depth of ocean (meters); margins of model domain must be zero:
  h_bo = zeros( lm + 2, mm + 2 );
  h_bo( 2 : end - 1, 2 : end - 1 ) = hfla;

  ndeg = get_nbr_deg_freedom( h_bo ); % Number of DEGrees of freedom.
  clear h_bo;

  grav = 9.8;                 % GRAVity (m/s2).
  cext = sqrt( grav * hfla ); % Celerity of surface gravity waves (m/s).

% Compute the first baroclinic Rossby radius of deformation,
%   sqrt( g' H_1 H_2 / (H_1+H_2) ) / f
  lros = grav * (rhon(2) - rhon(1)) / rhon(2);
  lros = sqrt( lros * topl(2) * hfla * (1. - topl(2)) ) / mean(fcor(:));
  disp([ 'Baroclinic Rossby radius = ' num2str( lros * 1.e-3 ) ' km.' ]);

% Write input files over disk. Before doing so, make sure that the directory
%   `outd' is clean (i.e., delete all input files inside `outd').

  if ( exist([ outd 'init.bin' ]) ); delete([ outd 'init.bin' ]); end
  if ( exist([ outd 'h_bo.bin' ]) ); delete([ outd 'h_bo.bin' ]); end
  if ( exist([ outd 'fcor.bin' ]) ); delete([ outd 'fcor.bin' ]); end
  if ( exist([ outd 'taus.bin' ]) ); delete([ outd 'taus.bin' ]); end
  if ( exist([ outd 'bodf.bin' ]) ); delete([ outd 'bodf.bin' ]); end
  if ( exist([ outd 'nudg.bin' ]) ); delete([ outd 'nudg.bin' ]); end
  if ( exist([ outd 'hdot.bin' ]) ); delete([ outd 'hdot.bin' ]); end
  if ( exist([ outd 'tide.bin' ]) ); delete([ outd 'tide.bin' ]); end

% Note that a 2-day ramp (dt_r=2.) is applied on the wind forcing to prevent
%   the inertial oscillations that accompany a sudden increase in winds
%   (e.g., Millot and Crepon 1981).

% print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, dt_s, ...
%               dt_o, dt_r, dt3d, bvis, dvis, bdrg, hmin, hsbl, hbbl, g_fb, ...
%               uadv, qdrg, ocrp, rsta, xper, yper, diag, tauw, idir, odir, ...
%               desc );
  print_params( lm,   mm,   nlay, ndeg, dl,   cext, 0.,   rhon, topl, 365., ...
                dt_o, 2.,   0.,   0.,   dvis, 0.,   hmin, hsbl, 100., 1.,   ...
                1.,   0.,   1.,   0,    0.,   0.,   1.,   [0,0],outd, outd, ...
                'Test-case: Double gyre ocean basin (Salmon 2002)' );

  [f_id, mesg] = fopen([ outd 'fcor.bin' ], 'w', 'ieee-le' );
  cont = fwrite( f_id, fcor,       'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont fcor;

  [f_id, mesg] = fopen([ outd 'taus.bin' ], 'w', 'ieee-le' );
  cont = fwrite( f_id, tauw,       'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

% See beom documentation on how to compile and execute the program.

  disp(['Now compile and execute the program in a different terminal. ' ...
        'Press a key when it is over.']);
  pause

  disp('Reading the output data, please wait...');

% Read the calculation's metadata:

  [taxi, h_0, f0, dl, rhon, desc] = get_metadata( strtrim( outd ) );
  nrec = length( taxi ); % Number of RECords in the output files.

% MasK for variables positioned at the `psi' point  (i.e., lower-left corner).
%   The psi points along the coastline are Excluded (i.e., set to zero).
  mkpe = ones( lm + 2, mm + 2, nlay );
  for ilay = 1 : nlay
    mkpe( 2 : end, 2 : end, ilay ) = mkpe( 2 : end,     2 : end,     ilay )...
                                  .* h_0(  1 : end - 1, 2 : end,     1    )...
                                  .* h_0(  2 : end,     2 : end,     1    )...
                                  .* h_0(  1 : end - 1, 1 : end - 1, 1    )...
                                  .* h_0(  2 : end,     1 : end - 1, 1    );
  end; clear ilay;
  mkpe( find( mkpe > 1.e-3 ) ) = 1.;
  mkpe(1, :, :) = 0.; mkpe(:, 1, :) = 0.; % Edges of model domain.

% MasK for variables positioned at the `eta' point (i.e., cell-center).
  mk_n = h_0; mk_n( find( mk_n > 1.e-3 ) ) = 1.;

% Time-average the Montgomery potential and later plot it:
  mmon = zeros( lm + 2, mm + 2, nlay ); % Mean Montgom.pot. / grav (m).

  if ( vide ); figure; colormap( jet ); end
  for irec = 1 : nrec - 1

%   Average temporally the thickness and Montgomery potential (divided by g):
    mont = get_field( 'mont', irec, outd );
    mmon = mmon + mont / nrec;

    if ( vide ) % Create a video of potential vorticity (layer 1).
      pvor = get_field( 'pvor', irec, outd );
      hand = imagesc( squeeze( pvor(2 : end - 1, 2 : end - 1, 1) )' );
      axis xy;
      set( gca, 'xtick', [], 'ytick', [] );
      ylabel([ 'Potential vorticity after ' ...
               num2str( round(taxi(irec)), '%04u' ) ' days' ]);
      axis image; caxis([ 0., 3.5e-6 ]); colorbar;
      print( '-dpng', [ '/tmp/frame' num2str( irec, '%04u' ) '.png' ] );
    end

%   If ebud=1, compute an energy budget. Otherwise, skip the rest of loop.
    if ( ~ ebud ); continue; end

%   Layers' thickness from current, past (-1), and future (+1) time records will
%     be referred to as: h_m1, hlay, h_p1 (respectively).
%   For kinetic energy:  k_m1, kine, k_p1 (respectively).

    if ( irec <= 2 )
      n = get_field( 'eta_', irec, outd );
      u = get_field( 'u___', irec, outd );
      v = get_field( 'v___', irec, outd );

      hlay = nan( lm + 2, mm + 2, nlay ); % Compute layers' thickness.
      for ilay = 1 : nlay - 1
        hlay(:,:,ilay) = h_0(:,:,ilay) + n(:,:,ilay) - n(:,:,ilay + 1);
      end; clear ilay;
      hlay(  :,:,nlay) = h_0(:,:,nlay) + n(:,:,nlay); clear n;

      u( find( isnan( u ) ) ) = 0.;
      v( find( isnan( v ) ) ) = 0.;

%     Eq.3.41 of Arakawa&Lamb 1981 https://doi.org/10.1175/MWR-D-16-0250.1
      kine = zeros( lm + 2, mm + 2, nlay );
      kine(  1 : end - 1, 1 : end - 1, : )           ...
        = u( 1 : end - 1, 1 : end - 1, : ).^2 * 0.25 ...
        + u( 2 : end,     1 : end - 1, : ).^2 * 0.25 ...
        + v( 1 : end - 1, 1 : end - 1, : ).^2 * 0.25 ...
        + v( 1 : end - 1, 2 : end,     : ).^2 * 0.25; clear u v;
      kine = kine .* hlay;
    end
    if ( irec == 1 )
      h_m1 = hlay;
      k_m1 = kine;
      dkin = zeros( nrec, nlay ); % Rate of change kin. energy (m^3 s^(-3)).
      dpot = zeros( nrec, nlay ); % Rate of change pot. energy (m^3 s^(-3)).
      diss = zeros( nrec, nlay ); % Dissipation by eddy visc.  (m^3 s^(-3)).
      wind = zeros( nrec, nlay ); % Wind's rate of work        (m^3 s^(-3)).
      continue;
    end

%   Compute thickness and kinetic energy for the next time record (+1):
    n = get_field( 'eta_', irec + 1, outd );
    u = get_field( 'u___', irec + 1, outd );
    v = get_field( 'v___', irec + 1, outd );

    h_p1 = nan( lm + 2, mm + 2, nlay );
    for ilay = 1 : nlay - 1
      h_p1(:,:,ilay) = h_0(:,:,ilay) + n(:,:,ilay) - n(:,:,ilay + 1);
    end; clear ilay;
    h_p1(  :,:,nlay) = h_0(:,:,nlay) + n(:,:,nlay);

    u( find( isnan( u ) ) ) = 0.;
    v( find( isnan( v ) ) ) = 0.;
    k_p1 = zeros( lm + 2, mm + 2, nlay );
    k_p1(  1 : end - 1, 1 : end - 1, : )           ...
      = u( 1 : end - 1, 1 : end - 1, : ).^2 * 0.25 ...
      + u( 2 : end,     1 : end - 1, : ).^2 * 0.25 ...
      + v( 1 : end - 1, 1 : end - 1, : ).^2 * 0.25 ...
      + v( 1 : end - 1, 2 : end,     : ).^2 * 0.25;
    k_p1 = k_p1 .* h_p1;

%   Compute rate of change of kinetic and potential energy by centered
%     finite difference, and then sum over the horizontal grid:
    wrk1 = (h_p1 - h_m1) / (2. * dt_o * 24. * 3600.) .* mont * grav;
    wrk1 = reshape( wrk1, [(lm + 2) * (mm + 2), nlay] );
    dpot( irec, : ) = nansum( wrk1 )'; clear wrk1;

    wrk1 = (k_p1 - k_m1) / (2. * dt_o * 24. * 3600.);
    wrk1 = reshape( wrk1, [(lm + 2) * (mm + 2), nlay] );
    dkin( irec, : ) = nansum( wrk1 )'; clear wrk1;

%   Compute dissipation due to sub-grid scale epipycnal viscosity `mu' (m2/s).

    dive = zeros( lm + 2, mm + 2, nlay ); % DIVErgence         (s^(-1)).
    rvor = zeros( lm + 2, mm + 2, nlay ); % Relative VORticity (s^(-1)).
    dive( 1 : end - 1, 1 : end - 1, :)      ...
      = u(2 : end,     1 : end - 1, :) / dl ...
      - u(1 : end - 1, 1 : end - 1, :) / dl ...
      + v(1 : end - 1, 2 : end,     :) / dl ...
      - v(1 : end - 1, 1 : end - 1, :) / dl;

    rvor( 2 : end,     2 : end,     :)      ...
      = u(2 : end,     1 : end - 1, :) / dl ...
      - u(2 : end,     2 : end,     :) / dl ...
      + v(2 : end,     2 : end,     :) / dl ...
      - v(1 : end - 1, 2 : end,     :) / dl;
    dive = dive .* mk_n; % Zero over land cells.
    rvor = rvor .* mkpe; % Zero along coastline.

    wrk1 = get_field( 'v_cc', irec + 1, outd ); % mu at cell-center.
    wrk2 = get_field( 'v_ll', irec + 1, outd ); % mu at lower-left corner.
    wrk1( find( isnan( wrk1 ) ) ) = 0.;
    wrk2( find( isnan( wrk2 ) ) ) = 0.;

    wrk1 = wrk1 .* dive; clear dive; % mu dive
    wrk2 = wrk2 .* rvor; clear rvor; % mu rvor

    wrk3 = zeros( lm + 2, mm + 2, nlay );
    wrk4 = zeros( lm + 2, mm + 2, nlay );
    wrk5 = zeros( lm + 2, mm + 2, nlay );
    wrk6 = zeros( lm + 2, mm + 2, nlay );

    wrk3(      2 : end,     :, :) ... %   d/dx (mu dive), at the u point.
      = ( wrk1(2 : end,     :, :) ...
        - wrk1(1 : end - 1, :, :) ) / dl;

    wrk4(      :, 1 : end - 1, :) ... % - d/dy (mu rvor), at the u point.
      = ( wrk2(:, 2 : end,     :) ...
        - wrk2(:, 1 : end - 1, :) ) / dl * (- 1.);

    wrk5(      1 : end - 1, :, :) ... %   d/dx (mu rvor), at the v point.
      = ( wrk2(2 : end,     :, :) ...
        - wrk2(1 : end - 1, :, :) ) / dl; clear wrk2;

    wrk6(      :, 2 : end,     :) ... %   d/dy (mu dive), at the v point.
      = ( wrk1(:, 2 : end,     :) ...
        - wrk1(:, 1 : end - 1, :) ) / dl; clear wrk1;

    wrk1 = h_p1; wrk1( find( isnan( wrk1 ) ) ) = 0.;
    hatu = wrk1(1 : end - 1, :, :) * 0.5 ...
         + wrk1(2 : end,     :, :) * 0.5; % h at the u point.
    hatv = wrk1(:, 1 : end - 1, :) * 0.5 ...
         + wrk1(:, 2 : end,     :) * 0.5; % h at the v point.
    clear wrk1;

    xdis = zeros( lm + 2, mm + 2, nlay );
    ydis = zeros( lm + 2, mm + 2, nlay );

    xdis(2 : end, :, :) = u(    2 : end, :, :) .* hatu ...
                       .* (wrk3(2 : end, :, :) + wrk4(2 : end, :, :));
    clear wrk3 wrk4 hatu;

    ydis(:, 2 : end, :) = v(    :, 2 : end, :) .* hatv ...
                       .* (wrk5(:, 2 : end, :) + wrk6(:, 2 : end, :));
    clear wrk5 wrk6 hatv v;

    xdis = reshape( xdis + ydis, [(lm + 2) * (mm + 2), nlay] );
    diss( irec + 1, : ) = nansum( xdis )'; clear xdis ydis;

%   Compute the wind's rate of work.
%   layt is an array used to distribute the wind stress across isopycnal layers
%     (see beom's documentation and `distribute_stress' inside private_mod.f95).
    layt = zeros( lm + 2, mm + 2, nlay );

    for ilay = 1 : nlay
      hcum = zeros( lm + 2, mm + 2 );
      sofr = sum( layt(:, :, 1 : ilay), 3 );
      for klay = 1 : ilay
        hcum = hcum + max( cat( 3, h_p1(:, :, klay) - 1.5 * (10. * hmin), ...
                                   zeros(lm + 2, mm + 2, 1) ), [], 3 );
      end; clear klay;
      layt(:, :, ilay) ...
        = min( cat( 3, hcum, hsbl * ones(lm + 2, mm + 2, 1) ), [], 3 ) ...
        / hsbl - sofr;
      layt(:, :, ilay) ...
        = max( cat( 3, layt(:, :, ilay), zeros(lm + 2, mm + 2, 1) ), [], 3 );
    end; clear ilay hcum sofr;
    layt( find( isnan( layt ) ) ) = 0.;

%   Zonal winds' rate of work: int[ rho^(-1) u . dtaux/dz ]dz, at the u points:
    wrk1 = zeros( lm + 2, mm + 2, nlay );
    for ilay = 1 : nlay
      wrk1(      1 : end - 1, :, ilay) ...
        = u(     1 : end - 1, :, ilay) ...
       .* ( layt(1 : end - 1, :, ilay) .* tauw(1 : end - 1, :, 1)   ...
          + layt(2 : end,     :, ilay) .* tauw(2 : end,     :, 1) ) ...
        * 0.5 / rhon( ilay );
    end; clear ilay u;
      
    wrk1 = reshape( wrk1, [(lm + 2) * (mm + 2), nlay] );
    wind( irec + 1, : ) = nansum( wrk1 )'; clear wrk1;

%   Update the arrays in preparation for the next iteration:
    h_m1 = hlay;
    k_m1 = kine;

    hlay = h_p1;
    kine = k_p1;
  end; clear irec n mont hlay;

% Plot the time-averaged Montgomery potential (divided by gravity) for layer 1:
  figure; colormap( jet );
  [ccon, hcon] = contourf( squeeze( mmon(2 : end - 1, 2 : end - 1, 1) )' );
  set( gca, 'xtick', [0 : 20 : lm + 2], ...
            'ytick', [0 : 20 : mm + 2] );
  set( hcon, 'linecolor', 'none' ); axis image; colorbar;

  if ( ebud )
    figure;
%   Budget of mechanical energy is: dkin+dpot = wind + diss + residual.
    resi = sum( dkin' + dpot' ) - sum( wind' ) - sum( diss' );
    hand = plot( taxi / 365.25, sum( dkin' + dpot' ), ...
                 taxi / 365.25, sum( wind'         ), ...
                 taxi / 365.25, sum( diss'         ), ...
                 taxi / 365.25, resi );
    xlabel('Years of spin-up'); grid on;
    ylabel('Rate of change in energy (m^3 s^{-3})');
    legend('d energy / dt', 'Winds', 'Dissipation', 'Residual');
  end

