
% Test-case for a slope current forced by local winds. The domain is a f-plane
%   in the Southern Hemisphere (f=constant<0) and represents an idealized
%   continental shelf-slope region. The domain is periodic in the along-slope
%   direction, closed at the onshelf (southern) boundary, and has a wage sponge
%   at the offshelf (northern) boundary. The slope current flows westward and is
%   forced by local easterly winds. The initial condition is the state of rest
%   (u=v=0,eta=0).
% See Stewart & Thompson (GRL, 2012) to learn more about the Antarctic slope
%   current and the role of easterly winds.

  clear;
  close all;
  more  off;

  lm   =   1;              % Grid points in along -flow direction.
  mm   = 400;              % Grid points in across-flow direction.
  dl   = 5.e3;             % Mesh size (m).
  f0   = - 1.e-4;          % Coriolis parameter (rad s^(-1)).
  hmax = 3000.;            % Maximum bottom depth over model domain (m).
  hshf = 1. / 6.;          % Depth of continental shelf (in fraction of hmax).
  y_cs = 0.25;             % Non-dim. position of continental slope.
% Path to input/output files (must end with character '/'):
  outd = '/tmp/';
  weak = 31.;              % Timescale for weakest nudging at boundary (days).
  npts = 15;               % Grid points in nudging zone (Modave et al 2010).
  grav = 9.8;              % Gravity (m s^(-2)).

% Assume a continuous profile of potential density of the form:
%   rho(z) = rho_bottom - exp( z / efol ) * drho (kg/m3),
%   where z is positive upward.
  nlay = 6;                   % Number of vertical layers.
  rbot = 1027.;               % Density of bottom layer (kg/m3).
  drho = 1.;                  % Top-to-bottom density contrast (kg/m3).
  efol = 1000.;               % e-folding scale for density profile (m).
  zcon = (- hmax : 1. : 0.)'; % Continuous profile "z" (m).
  rcon = rbot - exp( zcon / efol ) * drho;

% Discretize the density profile assuming density classes of width drho/nlay:

  topl = zeros( nlay, 1 ); % Top of layers (fraction of hmax).
  rhon = nan(   nlay, 1 ); % Density of layers (kg/m3).
  for ilay = 1 : nlay
    indx = find( rcon > min(rcon) + drho * (ilay - 1) / nlay ...
               & rcon < min(rcon) + drho *  ilay      / nlay );
    rhon( ilay ) = mean( rcon( indx ) );
    if ( ilay == 1 )
      continue
    end
    ztop = efol * log( (rbot - min(rcon(indx))) / drho );
    topl( ilay ) = - ztop / hmax;
    clear indx ztop;
  end
  clear drho efol rbot zcon rcon ilay;

% Define dimensional horizontal space with yy=0 at the continental slope:

  [yy, xx] = meshgrid( (1 : mm + 2) - 1.5, ...
                       (1 : lm + 2) - 1.5 );
  yy   = yy - yy( 1, round( y_cs * mm ) ); clear xx y_cs;
  yy   = yy * dl; % Meters.

% Bottom depth of grid (m); see Stewart & Thompson 2012.

  h_bo = hshf * hmax ...
       + (hmax - hshf * hmax) * (0.5 * tanh(yy / (10. * dl)) + 0.5);
  h_bo(  1, :) = 0.;
  h_bo(end, :) = 0.;
  h_bo(:, 1  ) = 0.;
  h_bo(:, end) = 0.;

% Redefine yy so that yy=0 at the onshelf boundary (the "boundary" is positioned
%   between grid cell 1 and grid cell 2).

  yy   = yy - mean( yy( 1, 1 : 2 )' ); % (m).

% Wind stress at the sea surface (Pascals); see Stewart & Thompson 2012.

  tacc =   0.20;  % Pa.
  tasf = - 0.05;  % Pa.
  y_s  = 200.e3;  % m.
  y_t  = 800.e3;  % m.
  l_y  = mm * dl; % m.

  taux = zeros( lm + 2, mm + 2    ); % Zonal component of stress.
  taus = zeros( lm + 2, mm + 2, 2 ); % Vector (taux,tauy).

  offs = find( yy >= y_t            ); % Stress off the shelf.
  cosl = find( yy <  y_t & yy > y_s ); % Stress on cont.slope.
  wrk1 = tacc * sin( pi * (yy - y_t) / (l_y - y_t) ).^2;
  wrk2 = tasf * sin( pi * (y_t - yy) / (y_t - y_s) ).^2;
  taux( offs ) = wrk1( offs ); clear offs wrk1;
  taux( cosl ) = wrk2( cosl ); clear cosl wrk2;
  clear tacc tasf y_s y_t l_y;

  taus(:,   :,   1) = taux; clear taux;

% Ensure that dtau/dx=0 and dtau/dy=0 at the domain's boundaries so that Ekman
%   pumping is exactly zero there:

  taus(1,   :,   :) = taus(2,       :,       :);
  taus(:,   1,   :) = taus(:,       2,       :);
  taus(end, :,   :) = taus(end - 1, :,       :);
  taus(:,   end, :) = taus(:,       end - 1, :);

% Compute the number or degrees of freedom (ndeg):

  ndeg = get_nbr_deg_freedom( h_bo );

% Celerity of surface gravity waves (m/s):

  cext = sqrt( grav * max( h_bo(:) ) );
  dt   = 0.5 * dl / cext; % Model timestep (seconds).

% It is a good idea to have a wave sponge somewhere in the domain, so that
%   gravity waves have an outlet. We use the northern boundary for this purpose.
% `nudg' is the rate at which model variables are relaxed (or "nudged") at the
%   open boundaries. `nudg' is non-dimensional and defined at the center of the
%   grid cells.
% `nudg' can be converted to a dimensional relaxation time-scale `tau' with:
%   tau = dt / nudg, where `dt' is the model timestep (in seconds).
% The default behavior of the code is to relax the model fields toward the
%   initial condition.
% `nudg' is the same for all the vertical layers but it is allowed to vary
%   between the 3 model fields (`eta', `u', `v'). `nudg' is typically assembled
%   one side at a time (east, north, west, south) with the final value of `nudg'
%   being the maximum of the four individual arrays (e.g., Lavelle & Thacker
%   2008).

  nort = zeros( lm + 2, mm + 2, 3 ); % `nudg' at the northern boundary.

% Define `widt' as the dimensional width of the nudging zones:
  widt = npts * dl; % Meters.

  for j = 1 : mm + 2 % Northern boundary.
%  `xpos' is the position within the nudging zone (in units of grid points).
    xpos        = j - 1.5 + npts - mm;
    xpos        = max([ xpos;        0.  ]);
    xpos        = min([ xpos; npts - 0.5 ]);
%   See Modave et al. (2010) regarding the following line:
    nort(:,j,:) = dt * cext / widt * xpos / (npts - xpos);
  end; clear j xpos;

% For each variable (eta,u,v),
%   take the maximum of eastern/western/northern/southern sides
%   (e.g., Fig.2, Lavelle & Thacker 2008).

  nudg = zeros( lm + 2, mm + 2, 3 ); % `3' represents eta, u, v.
  for ivar = 1 : size( nudg, 3 )
    temp = max( [reshape(nudg(:,:,ivar), [1, (lm + 2) * (mm + 2)]); ...
                 reshape(nort(:,:,ivar), [1, (lm + 2) * (mm + 2)])] );
    nudg(:,:,ivar) = reshape( temp', lm + 2, mm + 2 );
    clear temp;
  end; clear ivar nort;

% Write input files over disk. Before doing so, make sure that the directory
%   `outd' is clean (i.e., delete all input files inside `outd').

  if ( exist([ outd 'init.bin' ]) ); delete([ outd 'init.bin' ]); end
  if ( exist([ outd 'h_bo.bin' ]) ); delete([ outd 'h_bo.bin' ]); end
  if ( exist([ outd 'fcor.bin' ]) ); delete([ outd 'fcor.bin' ]); end
  if ( exist([ outd 'taus.bin' ]) ); delete([ outd 'taus.bin' ]); end
  if ( exist([ outd 'bodf.bin' ]) ); delete([ outd 'bodf.bin' ]); end
  if ( exist([ outd 'nudg.bin' ]) ); delete([ outd 'nudg.bin' ]); end
  if ( exist([ outd 'hdot.bin' ]) ); delete([ outd 'hdot.bin' ]); end
  if ( exist([ outd 'tide.bin' ]) ); delete([ outd 'tide.bin' ]); end

  [f_id, mesg] = fopen([ outd  'h_bo.bin' ], 'w', 'ieee-le' );
  cont = fwrite( f_id,   h_bo, 'real*4',      0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

  [f_id, mesg] = fopen([ outd  'taus.bin' ], 'w', 'ieee-le' );
  cont = fwrite( f_id,   taus, 'real*4',      0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cont;

  [f_id, mesg] = fopen([ outd  'nudg.bin' ], 'w', 'ieee-le' );
  cout = fwrite( f_id,   nudg, 'real*4',      0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg cout;

% print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, dt_s, ...
%               dt_o, dt_r, dt3d, bvis, dvis, bdrg, hmin, hsbl, hbbl, g_fb, ...
%               uadv, qdrg, ocrp, rsta, xper, yper, diag, tauw, idir, odir, ...
%               desc );
  print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, 20.,  ...
                10.,  1.,   0.1,  1.,   0.3, 1.1e-3,0.1,  100., 100., 1.,   ...
                1.,   0.,   1.,   0.,   1.,   0.,   0.,   [0,0],outd, outd, ...
                'Test-case: Slope current driven by Antarctic Easterlies' );

% See beom documentation on how to compile and execute the program.

  disp(['Now compile and execute the program in a different terminal. ' ...
        'Press a key when it is over.']);
  pause

  disp('Reading the output data, please wait...');

% Get results from the calculation. First, read the metadata:

  [taxi, h_0, f0, dl, rhon, desc] = get_metadata( strtrim( outd ) );
  nrec = length( taxi );     % Number of records in output files.
  lm   = size( h_0, 1 ) - 2; % Size of grid.
  mm   = size( h_0, 2 ) - 2;
  nlay = size( h_0, 3 );

  n0 = get_field( 'eta_', 1,    strtrim( outd ) ); % Initial condition.
  n  = get_field( 'eta_', nrec, strtrim( outd ) ); % Final   condition.
  u  = get_field( 'u___', nrec, strtrim( outd ) ); % Final   condition.

  ix_0 = ceil(0.5 * (lm + 2)); % Center of domain in zonal dimension.

% Layer's thickness (meters):
  hlay = nan( lm + 2, mm + 2, nlay );

% Vertical position of interfaces (meters):
  int0 = nan( lm + 2, mm + 2, nlay ); % Initial condition.
  inte = nan( lm + 2, mm + 2, nlay ); % Final   condition.

  hlay(:,:,nlay) = squeeze( h_0(:,:,nlay) + n(:,:,nlay) );
  for ilay = 1 : nlay
    int0(:,:,ilay) = h_bo - squeeze(sum(h_0(:,:,ilay : nlay), 3));
    inte(:,:,ilay) = h_bo - squeeze(sum(h_0(:,:,ilay : nlay), 3)) ...
                   -        squeeze(    n(  :,:,ilay       )    );
    if ( ilay == nlay )
      continue % Skip the next line if ilay==nlay.
    end
    hlay(:,:,ilay) = squeeze( h_0(:,:,ilay) + n(:,:,ilay) - n(:,:,ilay + 1) );
  end; clear ilay;

% Vertical position at mid-height within the layers (meters):
  midd = nan( lm + 2, mm + 2, nlay );

  midd(  :,:,nlay) = h_bo - 0.5 * squeeze( hlay(:,:,nlay) );
  for ilay = 1 : nlay - 1
    midd(:,:,ilay) = h_bo - 0.5 * squeeze( hlay(:,:,ilay) ) ...
                   - squeeze( sum( hlay(:,:,ilay + 1 : nlay), 3 ) );
  end; clear ilay;

% Extrapolate the velocities up to the surface and down to the seafloor:
  uext = cat( 3, u(   :,:,1), u,    u(   :,:,end) );
  mext = cat( 3, inte(:,:,1), midd, sum( h_0, 3 ) );
  yext = repmat( yy, [1, 1, nlay + 2] );

  figure;
  subplot(4,1,1);
    hand = plot( yy(   ix_0, 2 : end - 1) * 1.e-3, ...
        squeeze( taus( ix_0, 2 : end - 1, 1 ) ) );
    grid on; set( hand, 'linewidth', 3 );
    title('Zonal wind stress (Pascals)');
  subplot(4,1,2);
    hand = plot( yy( ix_0, 2 : end - 1) * 1.e-3, ...
        squeeze( n(  ix_0, 2 : end - 1, 1 ) ) );
    set( gca, 'ylim', [- 1.5,0.5] );
    grid on; set( hand, 'linewidth', 3 );
    xlabel('Meridional position y (km)');
    title([ 'Sea surface height (m, snapshot after ' ...
            num2str( round(taxi(nrec)) ) ' days)' ]);
  figure;
  colormap( jet );
    cint = ( - 0.55 : 0.025 : 0.55 );
    [ccon, hcon] = contourf( squeeze(yext( ix_0, 2 : end - 1, : )) * 1.e-3, ...
                             squeeze(mext( ix_0, 2 : end - 1, : )),         ...
                             squeeze(uext( ix_0, 2 : end - 1, : )), cint );
    set( gca, 'xlim', [0., 2000.], 'xtick', (0. : 500. : 2000.) );
    axis ij; caxis([ cint(1), cint(end) ]); colorbar;
    set( hcon, 'linecolor', 'none' );
    xlabel('Meridional position y (km)'); ylabel('Depth (m)');
    title([ 'Zonal velocity (m s^{-1}, snapshot after ' ...
            num2str( round(taxi(nrec)) ) ' days)' ]); clear cint ccon hcon;
  hold on;
    hand = plot( yy(   ix_0, 2 : end - 1) * 1.e-3, ...
        squeeze( int0( ix_0, 2 : end - 1, :) ),    ...
                 yy(   ix_0, 2 : end - 1) * 1.e-3, ...
        squeeze( inte( ix_0, 2 : end - 1, :) ),    ...
                 yy(   ix_0, 2 : end - 1) * 1.e-3, ...
                 h_bo( ix_0, 2 : end - 1)' );
    set( hand(1        : nlay    ), 'color', 0. * [1 1 1]                 );
    set( hand(nlay + 1 : nlay * 2), 'color', 1. * [1 1 1], 'linewidth', 2 );
    set( hand(end),                 'color', 0. * [1 1 1], 'linewidth', 2 );
  hold off;

