% -- External Function: fiel = get_field( vnam, irec, outd );
%
% Returns a modeled FIELd in its gridded format (i,j,nlay).
%
% Function arguments:
% `vnam' (input ): 4-characters string identifying NAMe of Variable desired:
%    'eta_' Interface/surface deviation from state of rest (m),
%    'u___' velocity component along grid's x axis (m/s),
%    'v___' velocity component along grid's y axis (m/s),
%    'pvor' Potential  VORticity (1/(m s)),
%    'mont' MONTgomery potential / gravity (m), includes Salmon's artificial
%           term if option `ocrp' is activated in shared_mod.f95.
%    'v_cc' Epipycnal sub-grid scale ("eddy") viscosity at Cell-Center (m2/s).
%    'v_ll' Epipycnal sub-grid scale ("eddy") viscosity at Lower-Left  (m2/s).
% `irec' (input ): integer, Must be >=0 and <= the nbr of entries in `time.txt'.
%    1 = Initial condition,
%    0 = Last RECord available in output files.
% `outd' (input ): string, the path to the OUTput Directory (must end with '/').
% `fiel' (output): float array, the modeled FIELd in gridded format (i,j,nlay).
function [fiel] = get_field( vnam, irec, outdir )

  vnam = strtrim( vnam );
  nvar = ['eta_'; ... %  1
          'u___'; ... %  2
          'v___'; ... %  3
          'pvor'; ... %  4
          'smcc'; ... %  5
          'bern'; ... %  6
          'bave'; ... %  7
          'v_ll'; ... %  8
          'layt'; ... %  9
          'layb'; ... % 10
          'rvor'; ... % 11
          'mont'; ... % 12
          'v_cc'];    % 13
  ivar = 0;
  for i = 1 : size(nvar, 1)
    if ( strcmp( vnam, nvar(i, :) ) )
      ivar = i;
    end
  end; clear i;
  if ( ivar == 0 )
    error(['Unknown variable ' vnam '. See help.']);
  end

% Read metadata.

  [fid, msg] = fopen( [deblank(outdir) 'param_basin.txt'], 'r' );
  if ( fid <= 0 )
    error(msg);
  end
  not_eof = 1;
  while ( not_eof )
    line = strtrim( fgetl(fid) );
    if ( strcmp( line(1 : 4), 'desc' ) )
      not_eof = 0;
    end
    eval( line );
  end; clear not_eof;
  fclose(fid); clear fid msg;

  taxi = load([deblank(outdir) 'time.txt']);
  nrec = length(taxi); clear taxi;
  if ( irec == 0 )
    irec = nrec;
  end

  [fid, msg] = fopen([deblank(outdir) 'grid.bin' ], 'r', 'ieee-le');
  [val, cnt] = fread(fid, Inf, 'int32', 0, 'ieee-le');
  pos_b = val(1 : round(length(val) / 5.));
  fclose(fid); clear fid msg cnt;

  if     ( ivar == 2 ) % Variable defined at `u' point (left   edge).
    mask = val(1 + 2 * round(length(val) / 5.) : 3 * round(length(val) / 5.));
  elseif ( ivar == 3 ) % Variable defined at `v' point (bottom edge).
    mask = val(1 + 3 * round(length(val) / 5.) : 4 * round(length(val) / 5.));
  elseif ( (ivar == 4 || ivar == 8 || ivar == 11) ) % `psi' point (lower-left).
    mask = val(1 + 4 * round(length(val) / 5.) : 5 * round(length(val) / 5.));
  else                 % Variable defined at center of grid cell.
    mask = val(1 +     round(length(val) / 5.) : 2 * round(length(val) / 5.));
  end
  clear val;

  mask( find( mask == 0 ) ) = nan;

  fiel = nan( lm + 2, mm + 2, nlay );

  [fid, msg] = fopen([deblank(outdir) nvar(ivar,:) '.bin'], 'r', 'ieee-le');
  if ( fid <= 0 )
    return;
  end

  if ( irec > 1 )
    fseek(fid, 4 * (irec(1) - 1) * length(pos_b(:)) * nlay, 'bof');
  end
  [vect, cnt] = fread(fid, [length(pos_b(:)), nlay], 'real*4', 0, 'ieee-le');
  for ilay = nlay : - 1 : 1
    vari           = nan( lm + 2, mm + 2 );
    vect(:,  ilay) = squeeze(vect(:,ilay)) .* mask(:);
    vari(pos_b   ) =         vect(:,ilay);
    fiel(:,:,ilay) =         vari(:,:   );
  end; clear ilay;
  fiel = squeeze( fiel );
  fclose( fid );

