
% Test-case for isopycnal outcropping over a seamount and sloppy coast.
%   Verifies that the initial state corresponds to the state of rest (du/dt=0).
%   See Salmon, R. (2002), J. Mar. Res., vol.60, p.605-638.

  clear;
  close all;
  more  off;

% Path to input/output files (must end with character '/'):
  outd   = '/tmp/';
  lx     = 600.e3;         % Length of domain in zonal direction (meters).
  dl     =  5.e3;          % Mesh size (meters).
  nlay   = 5;              % Number of layers.
  fcor   = + 1.e-4;        % Coriolis parameter (rad/s).
  if ( nlay == 1 )
    rhon = 1000.;          % Density of layers (kg/m3).
    topl = 0.;             % Top     of layers (fraction of Hmax).
  else
    rhon = 1000. + 30. * (0. : 1. / (nlay - 1) : 1.);
    topl = (0 : nlay - 1) ./ nlay;
  end
  grav   = 9.8;            % m/s2
  hmax   = 300.;           % Max basin depth before adding seamount (meters).
  lm     = round(lx / dl); % Grid size in zonal (x) direction.
  if ( mod(lm, 2) == 0 )
    lm = lm + 1;           % Make sure it is odd.
  end
  mm     = lm;             % For 3-D calculation.
% mm     = 1;              % For 2-D calculation (x-z plane).

  [yy, xx] = meshgrid( 1 : mm + 2, 1 : lm + 2 );
  xx = xx - mean(xx(:));
  yy = yy - mean(yy(:));
  xx = xx * dl;
  yy = yy * dl;

% Create a closed basin with depth increasing from zero at the edges to hmax
%   at the center. Then, add a Gaussian seamount at the center.
%   The steepest bottom slopes are found on the sides of the seamount.

  ix_0 = ceil(0.5 * (mm + 2)); % Center of grid.

  h_bo = sqrt(xx.^2 + yy.^2) / dl;
  h_bo = h_bo ./ max(h_bo(:, ix_0)) * hmax;
  h_bo = max(h_bo(:, ix_0)) - h_bo;
  smnt = hmax - exp(- (xx.^2 + yy.^2) ./ (50.e3)^2) * 0.75 * hmax;
  h_bo = min([h_bo(:)'; smnt(:)'])'; clear smnt;
  h_bo = reshape(h_bo, lm + 2, mm + 2);
  h_bo( find( h_bo < min(h_bo(:, ix_0)) ) ) = 0.;

% Bottom slope (non-dimensional).

  dhdx = zeros( lm + 2, mm + 2 );
  dhdy = zeros( lm + 2, mm + 2 );
  dhdx( 2 : end - 1, : ) = (h_bo( 3 : end, : ) - h_bo( 1 : end - 2, : )) / (2. * dl);
  dhdy( :, 2 : end - 1 ) = (h_bo( :, 3 : end ) - h_bo( :, 1 : end - 2 )) / (2. * dl);
  slop = sqrt( dhdx.^2 + dhdy.^2 );
  clear dhdx dhdy;

% hsal is the `Salmon's thickness' (in meters).
% `Baroclinic' version of Salmon's criterion (Eq.6.5, Salmon 2002 JMR):
%    hsal > (slope * dl) * (delta rho / rho_0)

  hsal = 1.0 * max( slop( find( h_bo > 1.e-3 ) ) ) * dl ...
       * (rhon( 2 ) - rhon( 1 )) / rhon( 2 );

% `hmin' is the minimum layer thickness allowed during a calculation.
%   By default the code assumes `hsal = 10 * hmin'.

  hmin = hsal / 10.;                     % Meters.

% There should be no `wetting-drying' in this test-case.
% Therefore, you must ensure that hlay(1) is always >> hsal by drying-up all
%   cells of insufficient depth. Note that the top layer has to accommodate all
%   the deeper layers that are outcropped. Layers that are outcropped have
%   a thickness approximately equal to `hsal'.

  dryd = 10. * hsal + (nlay - 1) * hsal; % Meters.

  h_bo( find( h_bo < dryd ) ) = 0.;      % Everything under `dryd' is dryied-up.

  h_bo(  1, :) = 0.;                     % Enforce dry margins.
  h_bo(end, :) = 0.;
  h_bo(:, end) = 0.;
  h_bo(:,   1) = 0.;

  cext = sqrt(grav * max(h_bo(:))); % Celerity of surface gravity waves (m/s).

  ndeg = get_nbr_deg_freedom( h_bo );

% Write input files over disk. Before doing so, make sure that the directory
%   `outd' is clean (i.e., delete all input files inside `outd').

  if ( exist([ outd 'init.bin' ]) ); delete([ outd 'init.bin' ]); end
  if ( exist([ outd 'h_bo.bin' ]) ); delete([ outd 'h_bo.bin' ]); end
  if ( exist([ outd 'fcor.bin' ]) ); delete([ outd 'fcor.bin' ]); end
  if ( exist([ outd 'taus.bin' ]) ); delete([ outd 'taus.bin' ]); end
  if ( exist([ outd 'bodf.bin' ]) ); delete([ outd 'bodf.bin' ]); end
  if ( exist([ outd 'nudg.bin' ]) ); delete([ outd 'nudg.bin' ]); end
  if ( exist([ outd 'hdot.bin' ]) ); delete([ outd 'hdot.bin' ]); end
  if ( exist([ outd 'tide.bin' ]) ); delete([ outd 'tide.bin' ]); end

  [fid, msg] = fopen([outd 'h_bo.bin'], 'w', 'ieee-le');
  cnt = fwrite(fid, h_bo, 'real*4', 0,       'ieee-le');
  fclose(fid); clear fid msg cnt;

% print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, dt_s, ...
%               dt_o, dt_r, dt3d, bvis, dvis, bdrg, hmin, hsbl, hbbl, g_fb, ...
%               uadv, qdrg, ocrp, rsta, xper, yper, diag, tauw, idir, odir, ...
%               desc );
  print_params( lm,   mm,   nlay, ndeg, dl,   cext, fcor, rhon, topl, 15.,  ...
                0.2,  0.,   0.,   0.,   0.,   0.,   hmin, 10.,  10.,  1.,   ...
                1.,   0.,   1.,   0.,   0.,   0.,   1.,   [0,0],outd, outd, ...
                'Test-case for state of rest allowing isopycnal outcrop' );

% See beom documentation on how to compile and execute the program.

  disp(['Now compile and execute the program in a different terminal. ' ...
        'Press a key when it is over.']);
  pause

  disp('Reading the output data, please wait...');

% Read metadata.

  [taxi, h_0, f0, dl, rhon, desc] = get_metadata( strtrim( outd ) );
  nrec = length( taxi );     % Nbr of records in output files.
  lm   = size( h_0, 1 ) - 2; % Size of grid.
  mm   = size( h_0, 2 ) - 2;
  nlay = size( h_0, 3 );

  kine = nan( nrec, nlay );  % Kinetic   energy    of layers.
  pens = nan( nrec, nlay );  % Potential enstrophy of layers.

  for irec = 1 : nrec
    eta  = get_field( 'eta_', irec, outd );
    u    = get_field( 'u___', irec, outd );
    v    = get_field( 'v___', irec, outd );
    pvor = get_field( 'pvor', irec, outd );

    if ( any( h_bo( 2 : lm + 1, 2 : mm + 1    ) > 0.5 ...
      & isnan( eta( 2 : lm + 1, 2 : mm + 1, 1 ) ) ) )
      if ( irec > 1 )
        eta  = get_field( 'eta_', irec - 1, outd );
        u    = get_field( 'u___', irec - 1, outd );
        v    = get_field( 'v___', irec - 1, outd );
        pvor = get_field( 'pvor', irec - 1, outd );
      end
      error('Error: Calculation ended early.');
    end

    zlay = nan( lm + 2, mm + 2, nlay ); % Distance from geoid z=0 (m).
    hlay = nan( lm + 2, mm + 2, nlay ); % Layers' thickness (m).

    for ilay = 1 : nlay
      zlay(:,:,ilay) = h_bo - squeeze(sum(h_0(:,:,ilay : nlay), 3)) ...
                     - eta(:,:,ilay);
    end; clear ilay;

    hlay(  :,:,nlay) = h_0(:,:,nlay) + eta(:,:,nlay);
    for ilay = nlay - 1 : - 1 : 1
      hlay(:,:,ilay) = h_0(:,:,ilay) + eta(:,:,ilay) - eta(:,:,ilay + 1);
    end; clear ilay;

    hatp = zeros(lm + 2, mm + 2, nlay); % Thickness at `psi' point (lower-left).
    hzer = hlay; mask = hlay;
    hzer( find(   isnan( hzer ) ) ) = 0.;
    mask( find( ~ isnan( mask ) ) ) = 1;
    mask( find(   isnan( mask ) ) ) = 0;
    mask(2 : end, 2 : end, :) =   mask(1 : end - 1, 1 : end - 1, :) ...
                                + mask(1 : end - 1, 2 : end,     :) ...
                                + mask(2 : end,     1 : end - 1, :) ...
                                + mask(2 : end,     2 : end,     :);
    hatp(2 : end, 2 : end, :) = ( hzer(1 : end - 1, 1 : end - 1, :) ...
                                + hzer(1 : end - 1, 2 : end,     :) ...
                                + hzer(2 : end,     1 : end - 1, :) ...
                                + hzer(2 : end,     2 : end,     :) ...
                                ) ...
                             ./   mask(2 : end, 2 : end, :);
    clear hzer mask;
    pens(  irec, :) = nansum( reshape( 0.5 * pvor.^2 .* hatp * 0.5, ...
                                       [(lm + 2) * (mm + 2), nlay] ), 1 );

    if ( mm == 1 )
      kine(irec, :) = nansum( reshape( 0.5 * hlay .* (u.^2       ), ...
                                       [(lm + 2) * (mm + 2), nlay] ), 1 );
    else
      kine(irec, :) = nansum( reshape( 0.5 * hlay .* (u.^2 + v.^2), ...
                                       [(lm + 2) * (mm + 2), nlay] ), 1 );
    end

    if ( irec == 1 || irec == nrec )
      figure;
      subplot(1,2,1);
      hand = plot( xx(:,ix_0) / 1.e3, [squeeze(zlay(:,ix_0,:)), h_bo(:,ix_0)] );
      legend( num2str( (1 : nlay)' ) );
      set( hand(end), 'color', [0 0 0] );
      xlabel('Distance from seamount (km)');
      ylabel('Depth (m)')
      title(['Position of layer interface after ' num2str(taxi(irec)) ' days']);
      axis ij;
      axis([min(xx(:)) / 1.e3, max(xx(:)) / 1.e3, - 1., max(h_bo(:))]);
  
      subplot(1,2,2);
      plot( xx(:,ix_0) / 1.e3, squeeze(hlay(:,ix_0,:)) );
      legend( num2str( (1 : nlay)' ) );
      xlabel('Distance from seamount (km)');
      ylabel('Thickness (m)');
      title(['Layer thickness after ' num2str(taxi(irec)) ' days']);
      axis([min(xx(:)) / 1.e3, max(xx(:)) / 1.e3, 0, max(h_0(:)) + 1.]);
      pause(0.1);
    end
  end; clear irec;

% Verify that kinetic energy and enstrophy do not grow over time
%   (a test of the Coriolis-advection term).

  figure;
  subplot(2,1,1);
    hand = plot( taxi, kine );
    legend( num2str( (1 : nlay)' ) );
    ylabel('m^3 s^{-2}');
    title('Timeseries of kinetic energy per layer');
    axis([0, ceil(taxi(end))]);
  subplot(2,1,2);
    hand = plot( taxi, pens - repmat( pens(1,:)', [1, length(taxi)] )' );
    legend( num2str( (1 : nlay)' ) );
    xlabel('Time (days)');
    ylabel('m^{-1} s^{-2}');
    title('Timeseries of changes in potential enstrophy of layers');
    axis([0, ceil(taxi(end))]);

