
% Test-case for a current forced remotely (i.e., generated outside of the model
%   domain). The domain is a f-plane in the Southern Hemisphere (f=constant<0)
%   and has no topography (for simplicity). The domain is open on all four sides
%   with the remote current (RC) flowing to the left ("westward", u<0) and with
%   a wave sponge on the "northern" side. The characteristics of the RC are
%   prescribed (and maintained over time) at the upstream boundary (eastern
%   side). The bottom depth is unusually shallow and the stratification is
%   unusually strong in order to minimize the computational cost of the
%   testcase. The testcase can be used as a starting point for a generic
%   continental shelf+slope domain with a slope current and/or coastal current
%   with user-defined characteristics.

  clear;
  close all;
  more  off;

  lm   = 200;            % Grid points in along -flow direction.
  mm   = 150;            % Grid points in across-flow direction.
  f0   = - 1.4e-4;       % Coriolis parameter (rad s^(-1)).
  hmax = 10.;            % Maximum depth of model domain (m).
  umax = 0.4;            % Maximum speed of RC (m s^(-1)).
  stdv = 2.;             % Width of RC (non-dim. by Rossby radius).
  y_sc = 0.5;            % Non-dim. position of RC (in cross-flow direction).
  nlay = 2;              % Number of vertical layers.
  topl = [   0.,   0.5]; % Top of layers (non-dimensional).
  rhon = [1000., 1030.]; % Density of layers (kg m^(-3)).
  weak = 31.;            % Timescale for weakest nudging at boundaries (days).
  npts = 15;             % Grid points in the nudging zones (Modave et al 2010).
% Path to input/output files (must end with character '/'):
  outd = '/tmp/';
  grav = 9.8;            % Gravity (m s^(-2)).

% First baroclinic Rossby radius of deformation `lros' (m):

  lros = sqrt( grav * (rhon(2) - rhon(1)) / rhon(2) * topl(2) * hmax ) ...
       / abs( f0 );

% Define a relatively coarse mesh where the radius of deformation is
%   represented by only a few grid points (2, in this case).

  dl   = lros / 2.;      % Mesh size (m).

% Define dimensional horizontal space with yy=0 at the location of the RC:

  [yy, xx] = meshgrid( (1 : mm + 2) - 1.5, ...
                       (1 : lm + 2) - 1.5 );
  yy   = yy - yy( 1, round( y_sc * mm ) ); clear y_sc;
  yy   = yy * dl; % Meters.
  xx   = xx * dl;

% Define a westward RC in the top layer (1).
% The RC has a Gaussian profile with a given standard deviation.

  n        = zeros( lm + 2, mm + 2, nlay );
  u        = zeros( lm + 2, mm + 2, nlay );
  v        = zeros( lm + 2, mm + 2, nlay );
  u(:,:,1) = umax * sign( f0 ) * exp( - 0.5 * yy.^2 / (stdv * lros)^2 );
  clear stdv;

% Assume (for simplicity) that the RC is barotropic:

  u(:, :, 2 : nlay) = u(:, :, 1);

% The RC is in geostrophic balance with the sea surface elevation (n(:,:,1)).
% Assume that `u' is related to `n' by a 2nd-order centered difference.

  for iy = 3 : mm + 2
    n(:, iy, 1) = n(:, iy - 2, 1) ...
                - u(:, iy - 1, 1) * (2. * dl) / grav * f0;
  end; clear iy;

% Add a constant so that the elevation is zero on average over the domain:

  ssh  = squeeze( n(:, :, 1) );
  n(:, :, 1) = ssh - mean( ssh(:) ); clear ssh;

% Compute the number or degrees of freedom (ndeg):

  h_bo = zeros( lm + 2, mm + 2 ); % Bottom depth of grid (m).
  h_bo( 2 : end - 1, 2 : end - 1 ) = hmax;
  ndeg = get_nbr_deg_freedom( h_bo ); clear h_bo;

% Celerity of surface gravity waves (m/s):

  cext = sqrt( grav * max( hmax ) );
  dt   = 0.5 * dl / cext; % Model timestep (seconds).

% `nudg' is the rate at which model variables are relaxed (or "nudged") at the
%   open boundaries. `nudg' is non-dimensional and defined at the center of the
%   grid cells.
% `nudg' can be converted to a dimensional relaxation time-scale `tau' with:
%   tau = dt / nudg, where `dt' is the model timestep (in seconds).
% The default behavior of the code is to relax the model fields toward the
%   initial condition (file `init.bin'; see below).
% `nudg' is the same for all the vertical layers but it is allowed to vary
%   between the 3 model fields (`eta', `u', `v'). `nudg' is typically assembled
%   one side at a time (east, north, west, south) with the final value of `nudg'
%   being the maximum of the four individual arrays (e.g., Lavelle & Thacker
%   2008).

  east = zeros( lm + 2, mm + 2, 3 ); % Non-dimensional.
  nort = zeros( lm + 2, mm + 2, 3 );
  west = zeros( lm + 2, mm + 2, 3 );
  sout = zeros( lm + 2, mm + 2, 3 );

% Re-define the variable `weak' as a non-dimensional nudging rate, and define
%   `widt' as the dimensional width of the nudging zones.

  weak = dt / (weak * 24. * 3600.); % Non-dimensional.
  widt = npts * dl;                 % Meters.

  for j = 1 : mm + 2 % Southern boundary.
% For this particular boundary, the only thing needed is a weak relaxation to
%   prevent long-term drifts in the model fields. The nudging rate decreases
%   linearly away from the boundary. Use the same rate for all variables (for
%   simplicity).

%  `xpos' is the position within the nudging zone (in units of grid points).
    xpos        = npts - (j - 1.5);
    xpos        = max([ xpos;        0.  ]);
    xpos        = min([ xpos; npts - 0.5 ]);
    sout(:,j,:) = weak * xpos / npts;
  end; clear j xpos;

  for i = 1 : lm + 2 % Western boundary (i.e., downstream of the RC).
                     % Same as for the southern boundary (i.e. weak relaxation).
    xpos = npts - (i - 1.5);
    xpos = max([ xpos;        0.  ]);
    xpos = min([ xpos; npts - 0.5 ]);
    west(i,:,:) = weak * xpos / npts;
  end; clear i xpos;

  for j = 1 : mm + 2 % Northern boundary.
% It is a good idea to have a wave sponge somewhere in the domain, so that
%   gravity waves have an outlet. We use the northern boundary for this purpose.
%   See Eq.29 of Modave et al. (2010).
    xpos        = j - 1.5 + npts - mm;
    xpos        = max([ xpos;        0.  ]);
    xpos        = min([ xpos; npts - 0.5 ]);
    nort(:,j,:) = dt * cext / widt * xpos / (npts - xpos);
  end; clear j xpos;

% For the eastern boundary (i.e., upstream of the RC), the same weak nudging
%   rate is used everywhere except within the RC. The nudging rate is increased
%   within the RC so that its characteristics (speed, width...) are faithfully
%   reproduced and maintained over time. The nudging rate is one at the center
%   of the RC (equivalent to a relaxation timescale tau=dt) and is gradually
%   tapered down toward `weak' in the cross-flow direction.

  tape = abs( squeeze( u(end, :, 1) ) )' / umax; % Tapering.

  for i = 1 : lm + 2
    xpos = i - 1.5 + npts - lm;
    xpos = max([ xpos;        0.  ]);
    xpos = min([ xpos; npts - 0.5 ]);

%   Strong nudging within the core of the RC:
    strg = tape * dt * cext / widt * xpos / (npts - xpos);

%   Use weak nudging away from the RC:
    strg( find( strg < weak * xpos / npts ) ) = weak * xpos / npts;

    east( i, :, : ) = repmat( strg, [1, 1, 3] );
  end; clear weak widt tape i xpos strg;

% For each variable (eta,u,v),
%   take the maximum of eastern/western/northern/southern sides
%   (e.g., Fig.2, Lavelle & Thacker 2008).

  nudg = zeros( lm + 2, mm + 2, 3 ); % `3' represents eta, u, v.
  for ivar = 1 : size( nudg, 3 )
    temp = max( [reshape(nudg(:,:,ivar), [1, (lm + 2) * (mm + 2)]); ...
                 reshape(east(:,:,ivar), [1, (lm + 2) * (mm + 2)]); ...
                 reshape(west(:,:,ivar), [1, (lm + 2) * (mm + 2)]); ...
                 reshape(nort(:,:,ivar), [1, (lm + 2) * (mm + 2)]); ...
                 reshape(sout(:,:,ivar), [1, (lm + 2) * (mm + 2)])] );
    nudg(:,:,ivar) = reshape( temp', lm + 2, mm + 2 );
    clear temp;
  end; clear ivar east west nort sout;

% Write input files over disk. Before doing so, make sure that the directory
%   `outd' is clean (i.e., delete all input files inside `outd').

  if ( exist([ outd 'init.bin' ]) ); delete([ outd 'init.bin' ]); end
  if ( exist([ outd 'h_bo.bin' ]) ); delete([ outd 'h_bo.bin' ]); end
  if ( exist([ outd 'fcor.bin' ]) ); delete([ outd 'fcor.bin' ]); end
  if ( exist([ outd 'taus.bin' ]) ); delete([ outd 'taus.bin' ]); end
  if ( exist([ outd 'bodf.bin' ]) ); delete([ outd 'bodf.bin' ]); end
  if ( exist([ outd 'nudg.bin' ]) ); delete([ outd 'nudg.bin' ]); end
  if ( exist([ outd 'hdot.bin' ]) ); delete([ outd 'hdot.bin' ]); end
  if ( exist([ outd 'tide.bin' ]) ); delete([ outd 'tide.bin' ]); end

  [f_id, mesg] = fopen([ outd  'init.bin' ],     'w', 'ieee-le' );
  coun = fwrite( f_id, cat(4, n, u, v), 'real*4', 0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg coun;

  [f_id, mesg] = fopen([ outd  'nudg.bin' ],     'w', 'ieee-le' );
  coun = fwrite( f_id,   nudg, 'real*4',          0,  'ieee-le' );
  fclose( f_id ); clear f_id mesg coun;

% print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, dt_s, ...
%               dt_o, dt_r, dt3d, bvis, dvis, bdrg, hmin, hsbl, hbbl, g_fb, ...
%               uadv, qdrg, ocrp, rsta, xper, yper, diag, tauw, idir, odir, ...
%               desc );
  print_params( lm,   mm,   nlay, ndeg, dl,   cext, f0,   rhon, topl, 10.,  ...
                1.,   0.,   0.1,  0.,   0.3,  0.,   0.1,  1.,   1.,   1.,   ...
                1.,   0.,   1.,   0.,   0.,   0.,   0.,   [0,0],outd, outd, ...
                'Test-case: Remotely-forced current' );

% See beom documentation on how to compile and execute the program.

  disp(['Now compile and execute the program in a different terminal. ' ...
        'Press a key when it is over.']);
  pause

  disp('Reading the output data, please wait...');

% Get results from the calculation. First, read the metadata:

  [taxi, h_0, f0, dl, rhon, desc] = get_metadata( strtrim( outd ) );
  nrec = length( taxi );     % Number of records in output files.
  lm   = size( h_0, 1 ) - 2; % Size of grid.
  mm   = size( h_0, 2 ) - 2;
  nlay = size( h_0, 3 );

  u0 = get_field( 'u___', 1,    strtrim( outd ) );
  u  = get_field( 'u___', nrec, strtrim( outd ) );

  figure;
  subplot(2,1,1);
    imagesc( [min(xx(:)), max(xx(:))] * 1.e-3, ...
             [min(yy(:)), max(yy(:))] * 1.e-3, ...
             squeeze( u0(:,:,1) )' );
    axis xy; colorbar;
    title([ 'Zonal velocity, day ' num2str( taxi( 1  ) ) ]);
    ylabel('Meridional distance (km)');
    axis image;
  subplot(2,1,2);
    imagesc( [min(xx(:)), max(xx(:))] * 1.e-3, ...
             [min(yy(:)), max(yy(:))] * 1.e-3, ...
             squeeze( u( :,:,1) )' );
    axis xy; colorbar;
    title([ 'Zonal velocity, day ' num2str( taxi(nrec) ) ]);
    xlabel('Zonal distance (km)');
    ylabel('Meridional distance (km)');
    axis image;

